package org.eclipse.jdt.internal.compiler.batch;

import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.zip.ZipFile;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileReader;
import org.eclipse.jdt.internal.compiler.env.NameEnvironmentAnswer;
import org.eclipse.jdt.core.compiler.CharOperation;
import java.util.HashSet;
import org.eclipse.jdt.internal.compiler.util.Util;
import java.io.File;
import java.util.Map;
import org.eclipse.jdt.internal.compiler.env.AccessRuleSet;
import java.io.IOException;
import java.util.Set;
import org.eclipse.jdt.internal.compiler.util.SuffixConstants;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;

public class FileSystem implements INameEnvironment, SuffixConstants
{
    protected FileSystem.Classpath[] classpaths;
    Set knownFileNames;
    protected boolean annotationsFromClasspath;
    
    public FileSystem(String[] classpathNames, String[] initialFileNames, String encoding) {
        int classpathSize = classpathNames.length;
        this.classpaths = new FileSystem.Classpath[classpathSize];
        int counter = 0;
        for (int i = 0; i < classpathSize; ++i) {
            FileSystem.Classpath classpath = getClasspath(classpathNames[i], encoding, null, null);
            try {
                classpath.initialize();
                this.classpaths[counter++] = classpath;
            }
            catch (IOException ex) {}
        }
        if (counter != classpathSize) {
            System.arraycopy(this.classpaths, 0, this.classpaths = new FileSystem.Classpath[counter], 0, counter);
        }
        this.initializeKnownFileNames(initialFileNames);
    }
    
    protected FileSystem(FileSystem.Classpath[] paths, String[] initialFileNames, boolean annotationsFromClasspath) {
        int length = paths.length;
        int counter = 0;
        this.classpaths = new FileSystem.Classpath[length];
        for (FileSystem.Classpath classpath : paths) {
            try {
                classpath.initialize();
                this.classpaths[counter++] = classpath;
            }
            catch (IOException ex) {}
        }
        if (counter != length) {
            System.arraycopy(this.classpaths, 0, this.classpaths = new FileSystem.Classpath[counter], 0, counter);
        }
        this.initializeKnownFileNames(initialFileNames);
        this.annotationsFromClasspath = annotationsFromClasspath;
    }
    
    public static FileSystem.Classpath getClasspath(String classpathName, String encoding, AccessRuleSet accessRuleSet) {
        return getClasspath(classpathName, encoding, false, accessRuleSet, null, null);
    }
    
    public static FileSystem.Classpath getClasspath(String classpathName, String encoding, AccessRuleSet accessRuleSet, Map options) {
        return getClasspath(classpathName, encoding, false, accessRuleSet, null, options);
    }
    
    public static FileSystem.Classpath getClasspath(String classpathName, String encoding, boolean isSourceOnly, AccessRuleSet accessRuleSet, String destinationPath, Map options) {
        FileSystem.Classpath result = null;
        File file = new File(convertPathSeparators(classpathName));
        if (file.isDirectory()) {
            if (file.exists()) {
                result = new ClasspathDirectory(file, encoding, isSourceOnly ? 1 : 3, accessRuleSet, (destinationPath == null || destinationPath == "none") ? destinationPath : convertPathSeparators(destinationPath), options);
            }
        }
        else if (Util.isPotentialZipArchive(classpathName)) {
            if (isSourceOnly) {
                result = new ClasspathSourceJar(file, true, accessRuleSet, encoding, (destinationPath == null || destinationPath == "none") ? destinationPath : convertPathSeparators(destinationPath));
            }
            else if (destinationPath == null) {
                result = new ClasspathJar(file, true, accessRuleSet, null);
            }
        }
        return result;
    }
    
    private void initializeKnownFileNames(String[] initialFileNames) {
        if (initialFileNames == null) {
            this.knownFileNames = new HashSet(0);
            return;
        }
        this.knownFileNames = new HashSet(initialFileNames.length * 2);
        int i = initialFileNames.length;
        while (--i >= 0) {
            File compilationUnitFile = new File(initialFileNames[i]);
            char[] fileName = null;
            try {
                fileName = compilationUnitFile.getCanonicalPath().toCharArray();
            }
            catch (IOException ex) {
                continue;
            }
            char[] matchingPathName = null;
            int lastIndexOf = CharOperation.lastIndexOf('.', fileName);
            if (lastIndexOf != -1) {
                fileName = CharOperation.subarray(fileName, 0, lastIndexOf);
            }
            CharOperation.replace(fileName, '\\', '/');
            boolean globalPathMatches = false;
            for (int j = 0, max = this.classpaths.length; j < max; ++j) {
                char[] matchCandidate = this.classpaths[j].normalizedPath();
                boolean currentPathMatch = false;
                if (this.classpaths[j] instanceof ClasspathDirectory && CharOperation.prefixEquals(matchCandidate, fileName)) {
                    currentPathMatch = true;
                    if (matchingPathName == null) {
                        matchingPathName = matchCandidate;
                    }
                    else if (currentPathMatch) {
                        if (matchCandidate.length > matchingPathName.length) {
                            matchingPathName = matchCandidate;
                        }
                    }
                    else if (!globalPathMatches && matchCandidate.length < matchingPathName.length) {
                        matchingPathName = matchCandidate;
                    }
                    if (currentPathMatch) {
                        globalPathMatches = true;
                    }
                }
            }
            if (matchingPathName == null) {
                this.knownFileNames.add(new String(fileName));
            }
            else {
                this.knownFileNames.add(new String(CharOperation.subarray(fileName, matchingPathName.length, fileName.length)));
            }
            matchingPathName = null;
        }
        return;
    }
    
    public void cleanup() {
        for (int i = 0, max = this.classpaths.length; i < max; ++i) {
            this.classpaths[i].reset();
        }
    }
    
    private static String convertPathSeparators(String path) {
        return (File.separatorChar == '/') ? path.replace('\\', '/') : path.replace('/', '\\');
    }
    
    private NameEnvironmentAnswer findClass(String qualifiedTypeName, char[] typeName, boolean asBinaryOnly) {
        NameEnvironmentAnswer answer = this.internalFindClass(qualifiedTypeName, typeName, asBinaryOnly);
        if (this.annotationsFromClasspath && answer != null && answer.getBinaryType() instanceof ClassFileReader) {
            for (int i = 0, length = this.classpaths.length; i < length; ++i) {
                FileSystem.Classpath classpathEntry = this.classpaths[i];
                if (classpathEntry.hasAnnotationFileFor(qualifiedTypeName)) {
                    ZipFile zip = (classpathEntry instanceof ClasspathJar) ? ((ClasspathJar)classpathEntry).zipFile : null;
                    try {
                        ((ClassFileReader)answer.getBinaryType()).setExternalAnnotationProvider(classpathEntry.getPath(), qualifiedTypeName, zip, null);
                        break;
                    }
                    catch (IOException ex) {}
                }
            }
        }
        return answer;
    }
    
    private NameEnvironmentAnswer internalFindClass(String qualifiedTypeName, char[] typeName, boolean asBinaryOnly) {
        if (this.knownFileNames.contains(qualifiedTypeName)) {
            return null;
        }
        String qualifiedBinaryFileName = qualifiedTypeName + ".class";
        String qualifiedPackageName = (qualifiedTypeName.length() == typeName.length) ? Util.EMPTY_STRING : qualifiedBinaryFileName.substring(0, qualifiedTypeName.length() - typeName.length - 1);
        String qp2 = (File.separatorChar == '/') ? qualifiedPackageName : qualifiedPackageName.replace('/', File.separatorChar);
        NameEnvironmentAnswer suggestedAnswer = null;
        if (qualifiedPackageName == qp2) {
            for (int j = 0, length2 = this.classpaths.length; j < length2; ++j) {
                NameEnvironmentAnswer answer2 = this.classpaths[j].findClass(typeName, qualifiedPackageName, qualifiedBinaryFileName, asBinaryOnly);
                if (answer2 != null) {
                    if (!answer2.ignoreIfBetter()) {
                        if (answer2.isBetter(suggestedAnswer)) {
                            return answer2;
                        }
                    }
                    else if (answer2.isBetter(suggestedAnswer)) {
                        suggestedAnswer = answer2;
                    }
                }
            }
        }
        else {
            String qb2 = qualifiedBinaryFileName.replace('/', File.separatorChar);
            for (int i = 0, length = this.classpaths.length; i < length; ++i) {
                FileSystem.Classpath p = this.classpaths[i];
                NameEnvironmentAnswer answer = (p instanceof ClasspathJar) ? p.findClass(typeName, qualifiedPackageName, qualifiedBinaryFileName, asBinaryOnly) : p.findClass(typeName, qp2, qb2, asBinaryOnly);
                if (answer != null) {
                    if (!answer.ignoreIfBetter()) {
                        if (answer.isBetter(suggestedAnswer)) {
                            return answer;
                        }
                    }
                    else if (answer.isBetter(suggestedAnswer)) {
                        suggestedAnswer = answer;
                    }
                }
            }
        }
        if (suggestedAnswer != null) {
            return suggestedAnswer;
        }
        return null;
    }
    
    public NameEnvironmentAnswer findType(char[][] compoundName) {
        if (compoundName != null) {
            return this.findClass(new String(CharOperation.concatWith(compoundName, '/')), compoundName[compoundName.length - 1], false);
        }
        return null;
    }
    
    public char[][][] findTypeNames(char[][] packageName) {
        char[][][] result = null;
        if (packageName != null) {
            String qualifiedPackageName = new String(CharOperation.concatWith(packageName, '/'));
            String qualifiedPackageName2 = (File.separatorChar == '/') ? qualifiedPackageName : qualifiedPackageName.replace('/', File.separatorChar);
            if (qualifiedPackageName == qualifiedPackageName2) {
                for (int i = 0, length = this.classpaths.length; i < length; ++i) {
                    char[][][] answers2 = this.classpaths[i].findTypeNames(qualifiedPackageName);
                    if (answers2 != null) {
                        if (result == null) {
                            result = answers2;
                        }
                        else {
                            int resultLength2 = result.length;
                            int answersLength2 = answers2.length;
                            System.arraycopy(result, 0, result = new char[answersLength2 + resultLength2][][], 0, resultLength2);
                            System.arraycopy(answers2, 0, result, resultLength2, answersLength2);
                        }
                    }
                }
            }
            else {
                for (int i = 0, length = this.classpaths.length; i < length; ++i) {
                    FileSystem.Classpath p = this.classpaths[i];
                    char[][][] answers = (p instanceof ClasspathJar) ? p.findTypeNames(qualifiedPackageName) : p.findTypeNames(qualifiedPackageName2);
                    if (answers != null) {
                        if (result == null) {
                            result = answers;
                        }
                        else {
                            int resultLength = result.length;
                            int answersLength = answers.length;
                            System.arraycopy(result, 0, result = new char[answersLength + resultLength][][], 0, resultLength);
                            System.arraycopy(answers, 0, result, resultLength, answersLength);
                        }
                    }
                }
            }
        }
        return result;
    }
    
    public NameEnvironmentAnswer findType(char[][] compoundName, boolean asBinaryOnly) {
        if (compoundName != null) {
            return this.findClass(new String(CharOperation.concatWith(compoundName, '/')), compoundName[compoundName.length - 1], asBinaryOnly);
        }
        return null;
    }
    
    public NameEnvironmentAnswer findType(char[] typeName, char[][] packageName) {
        if (typeName != null) {
            return this.findClass(new String(CharOperation.concatWith(packageName, typeName, '/')), typeName, false);
        }
        return null;
    }
    
    public boolean isPackage(char[][] compoundName, char[] packageName) {
        String qualifiedPackageName = new String(CharOperation.concatWith(compoundName, packageName, '/'));
        String qp2 = (File.separatorChar == '/') ? qualifiedPackageName : qualifiedPackageName.replace('/', File.separatorChar);
        if (qualifiedPackageName == qp2) {
            int i = 0;
            int length = this.classpaths.length;
            while (i < length) {
                if (this.classpaths[i].isPackage(qualifiedPackageName)) {
                    return true;
                }
                ++i;
            }
        }
        else {
            int i = 0;
            int length = this.classpaths.length;
            while (i < length) {
                FileSystem.Classpath p = this.classpaths[i];
                if (!((p instanceof ClasspathJar) ? (!p.isPackage(qualifiedPackageName)) : (!p.isPackage(qp2)))) {
                    return true;
                }
                ++i;
            }
        }
        return false;
    }
    
    public static class ClasspathNormalizer
    {
        public static ArrayList normalize(ArrayList classpaths) {
            ArrayList<FileSystem.Classpath> normalizedClasspath = new ArrayList();
            HashSet<FileSystem.Classpath> cache = new HashSet();
            for (FileSystem.Classpath classpath : classpaths) {
                if (!cache.contains(classpath)) {
                    normalizedClasspath.add(classpath);
                    cache.add(classpath);
                }
                else {
                    continue;
                }
            }
            return normalizedClasspath;
        }
    }
    
    public interface Classpath
    {
        public abstract char[][][] findTypeNames(String p0);
        
        public abstract NameEnvironmentAnswer findClass(char[] p0, String p1, String p2);
        
        public abstract NameEnvironmentAnswer findClass(char[] p0, String p1, String p2, boolean p3);
        
        public abstract boolean isPackage(String p0);
        
        public abstract List fetchLinkedJars(FileSystem.ClasspathSectionProblemReporter p0);
        
        public abstract void reset();
        
        public abstract char[] normalizedPath();
        
        public abstract String getPath();
        
        public abstract void initialize() throws IOException;
        
        public abstract boolean hasAnnotationFileFor(String p0);
    }
    
    public interface ClasspathSectionProblemReporter
    {
        public abstract void invalidClasspathSection(String p0);
        
        public abstract void multipleClasspathSections(String p0);
    }
}
